/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2010.                   *
*                  Copyright (C) Svante Signell, 2013.                    *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/


/* scm_rights+creds_recv.c

   Used in conjunction with scm_rights+creds_send.c to demonstrate
   passing of file descriptors and process credentials via a UNIX
   domain socket.

   This program receives file desciptors and credentials sent to a
   UNIX domain socket.

   Usage is as shown in the usageErr() call below.

   File descriptors and process credentials can exchanged over stream
   or datagram sockets. This program uses stream sockets by default;
   the "-d" command-line option specifies that datagram sockets should
   be used instead.

   Modified for kFreeBSD and Hurd by Svante Signell April-Aug 2013.
   Added -z option, send a zero byte for credentials with Hurd and
   kFreeBSD New file: scm_rights+creds_recv.c, Svante Signell Nov 2013
   based on modified versions of scm_rights_recv.c and
   scm_creds_recv.c.

   This program is no longer Linux-specific.
*/
#include "scm_rights+creds.h"

#ifndef SCM_CREDENTIALS
#define SCM_CREDENTIALS SCM_CREDS;
#endif
#define BUF_SIZE 100

#ifndef NRIGHTS
#define NRIGHTS 1
#endif

#ifndef NCREDS
#define NCREDS 0
#endif

int
main(int argc, char *argv[])
{
  struct msghdr msgh;
  struct iovec iov;
  int data;
  int lfd, sfd, opt;

#if NRIGHTS !=0
  int *fdptr = NULL, *end = NULL;
#endif

#if NCREDS != 0
#if defined (__linux__)
  struct ucred ucred, *ucredp;
  int sz =  sizeof(struct ucred);
#else
  struct cmsgcred *ucredp;
  int sz =  sizeof(struct cmsgcred);
#endif
#else
  int sz = 0;
#endif

  /* Space large enough to hold NCREDS ucred structures + NRIGHTS FDs */
  /* FIXME: Cases 0 does not work */

  int SW, cspc, cspc1, cspc2, cspc3;
  if (NRIGHTS == 0 && !NCREDS) SW = 0;
  if (NRIGHTS == 1 && !NCREDS) SW = 1;
  if (NRIGHTS == 2 && !NCREDS) SW = 2;
  if (NRIGHTS == 3 && !NCREDS) SW = 3;
  if (NRIGHTS == 0 && NCREDS) SW = 4;
  if (NRIGHTS == 1 && NCREDS) SW = 5;
  if (NRIGHTS == 2 && NCREDS) SW = 6;
  if (NRIGHTS == 3 && NCREDS) SW = 7;

  cspc1 = CMSG_SPACE(sizeof(int));
  cspc2 = CMSG_SPACE(sz);
  cspc3 = CMSG_SPACE(2 * sizeof(int));
  switch (SW)
    {
    case 0:
      cspc = CMSG_SPACE(sizeof (int));
      fprintf(stderr,"Case 1: cspc = %c [char]\n", cspc);
      break;
    case 1:
      cspc = cspc1;
      fprintf(stderr,"Case 1: cspc = %d\n", cspc);
      break;
    case 2:
      cspc = cspc3;
      fprintf(stderr,"Case 2: cspc = %d\n", cspc);
      break;
    case 3:
      cspc = cspc1 + cspc3;
      fprintf(stderr,"Case 3: cspc = %d\n", cspc);
      break;
    case 4:
      cspc = cspc2;
      fprintf(stderr,"Case 4: cspc = %d\n", cspc);
      break;
    case 5:
      cspc = cspc1 + cspc2;
      fprintf(stderr,"Case 5: cspc = %d\n", cspc);
      break;
    case 6:
      cspc = cspc2 + cspc3;
      fprintf(stderr,"Case 6: cspc = %d\n", cspc);
      break;
    case 7:
      cspc = cspc1 + cspc2 + cspc3;
      fprintf(stderr,"Case 7: cspc = %d\n", cspc);
      break;
    default:
      break;
    }

#if NRIGHTS !=0 || NCREDS !=0
  union {
    struct cmsghdr cmh;
    char control[cspc];
  } control_un;
  struct cmsghdr *cmhp = NULL;
#endif

#if defined (__linux__) && NCREDS != 0
  int optval;
  socklen_t len;
#endif

  ssize_t nr;
  Boolean useDatagramSocket;

    /* Parse command-line arguments */
    useDatagramSocket = FALSE;

    while ((opt = getopt(argc, argv, "d")) != -1) {
        switch (opt) {
        case 'd':
            useDatagramSocket = TRUE;
            break;

        default:
            usageErr("%s [-d]\n"
                     "        -d    use datagram socket\n", argv[0]);
        }
    }

    /* Create socket bound to well-known address */
    if (remove(SOCK_PATH) == -1 && errno != ENOENT)
        errExit("remove-%s", SOCK_PATH);

    if (useDatagramSocket) {
        fprintf(stderr, "Receiving via datagram socket\n");
        sfd = unixBind(SOCK_PATH, SOCK_DGRAM);
        if (sfd == -1)
            errExit("unixBind");

    } else {
        fprintf(stderr, "Receiving via stream socket\n");
        lfd = unixListen(SOCK_PATH, 5);
        if (lfd == -1)
            errExit("unixListen");

        sfd = accept(lfd, NULL, NULL);
        if (sfd == -1)
            errExit("accept");
    }

    /* We must set the SO_PASSCRED socket option in order to receive
       credentials */
#if defined(__linux__) && NCREDS != 0
    /* SO_PASSCRED not defined on GNU */
    optval = 1;
    if (setsockopt(sfd, SOL_SOCKET, SO_PASSCRED, &optval, sizeof(optval)) == -1)
      errExit("setsockopt");
#endif

#if NRIGHTS !=0 || NCREDS !=0
/* FIXME: To remove ?? */
#if 0
    /* Set 'control_un' to describe ancillary data that we want to receive */
    control_un.cmh.cmsg_len = CMSG_LEN(sizeof(int));
    control_un.cmh.cmsg_level = SOL_SOCKET;
    control_un.cmh.cmsg_type = SCM_RIGHTS;
#endif

    /* Set 'msgh' fields to describe 'control_un' */
    msgh.msg_control = control_un.control;
    msgh.msg_controllen = sizeof(control_un.control);
    fprintf(stderr, "scm_rights_recv.c:msgh.msg_controllen = %d\n", (int)msgh.msg_controllen);
#endif

    /* Set fields of 'msgh' to point to buffer used to receive (real)
       data read by recvmsg() */
    msgh.msg_iov = &iov;
    msgh.msg_iovlen = 1;
    iov.iov_base = &data;
    iov.iov_len = sizeof(int);

    msgh.msg_name = NULL;               /* We don't need address of peer */
    msgh.msg_namelen = 0;

    /* Receive real plus ancillary data */
    nr = recvmsg(sfd, &msgh, 0);
    if (nr == -1)
        errExit("recvmsg");
    fprintf(stderr, "recvmsg() returned %ld\n", (long) nr);
    
    if (nr > 0)
      fprintf(stderr, "Received data = %d\n", data);

#if NRIGHTS !=0 || NCREDS !=0
    /* Get the received file descriptor (which is typically a different
       file descriptor number than was used in the sending process) */
    int i = 0;
    for (cmhp = CMSG_FIRSTHDR(&msgh);
	 cmhp != NULL;
	 cmhp = CMSG_NXTHDR(&msgh, cmhp))
      {
	printf("cmhp->cmsg_len[%d] = %d\n", i, (int)cmhp->cmsg_len);
	i++;
	if (cmhp == NULL)
	  fatal("bad cmsg header cmhp == NULL");
	if (cmhp->cmsg_level != SOL_SOCKET)
	fatal("cmsg_level != SOL_SOCKET");

#if defined (__GNU__) || defined (__FreeBSD_kernel__)
	if (cmhp->cmsg_type != SCM_CREDS && cmhp->cmsg_type != SCM_RIGHTS)
	  fatal("cmsg_type != SCM_CREDS && cmsg_type != SCM_RIGHTS");
#else
	if (cmhp->cmsg_type != SCM_CREDENTIALS && cmhp->cmsg_type != SCM_RIGHTS)
	  fatal("cmsg_type != SCM_CREDENTIALS && cmsg_type != SCM_RIGHTS");
#endif

#if NCREDS != 0
#if defined (__GNU__) || defined (__FreeBSD_kernel__)
      if (cmhp->cmsg_type == SCM_CREDS)
	{
	  if (cmhp->cmsg_len != CMSG_LEN(sizeof(struct cmsgcred)))
	    fatal("bad cmsg message length cmhp->cmsg_len=%d, "
		  "CMSG_LEN(sizeof(struct cmsgcred))=%d",
		  cmhp->cmsg_len, CMSG_LEN(sizeof(struct cmsgcred)));
	  ucredp = (struct cmsgcred *) CMSG_DATA(cmhp);
	  printf("Received credentials pid=%ld, uid=%ld, euid=%ld, gid=%ld\n",
		 (long) ucredp->cmcred_pid, (long) ucredp->cmcred_uid, 
		 (long) ucredp->cmcred_euid, (long) ucredp->cmcred_gid);
	  for (i=0; i<ucredp->cmcred_ngroups; i++)
	    printf("ucredp->cmcred_groups[%d]=%d\n", i, ucredp->cmcred_groups[i]);
	  continue;
	}
#else /* Linux */
      if (cmhp->cmsg_type == SCM_CREDENTIALS)
	{
	  if (cmhp->cmsg_len != CMSG_LEN(sizeof(struct ucred)))
	    fatal("bad message length cmhp->cmsg_len=%d, "
		  "CMSG_LEN(sizeof(struct ucred))=%d",
		  cmhp->cmsg_len, CMSG_LEN(sizeof(struct ucred)));
	  ucredp = (struct ucred *) CMSG_DATA(cmhp);
	  printf("Received credentials: pid=%ld, uid=%ld, gid=%ld\n",
		 (long) ucredp->pid, (long) ucredp->uid, (long) ucredp->gid);
	  continue;
	}
#endif /* GNU+kFreeBSD */
#endif /* NCREDS */

      /* FIXME: To remove ?? */
#if 0
      if (cmhp->cmsg_len != CMSG_LEN(sizeof(int)) ||
	  cmhp->cmsg_len != CMSG_LEN(2*sizeof(int)))
        fatal("bad cmsg message length");
#endif

#if NRIGHTS !=0
      if (cmhp->cmsg_type != SCM_RIGHTS)
	fatal("cmsg_type != SCM_RIGHTS");
      fdptr = (int *) CMSG_DATA(cmhp);
      end = (int *)((char *)cmhp + cmhp->cmsg_len);

      /* Having obtained the file descriptor, read the file's contents and
	 print them on standard output */
      int i = 0;
      for (i=0; fdptr+i < end; i++)
	{
	  printf("Received fd[%d] = %d\n", i, fdptr[i]);

	  char buf[BUF_SIZE];
	  ssize_t numRead;

	  numRead = read(fdptr[i], buf, BUF_SIZE);
	  if (numRead == -1)
	    errExit("read");

	  if (numRead == 0)
	    break;
	  printf("Contents:\n");
	  write(STDOUT_FILENO, buf, numRead);
	  printf("\n");
	}
      continue;
#endif /* NRIGHTS */
      } /* for (cmsg=...) */
#endif /* NRIGHTS !=0 || NCREDS !=0 */

#if defined(__linux__) && NCREDS != 0
  /* The Linux-specific, read-only SO_PEERCRED socket option returns
     credential information about the peer, as described in socket(7) */
  len = sizeof(struct ucred);
  if (getsockopt(sfd, SOL_SOCKET, SO_PEERCRED, &ucred, &len) == -1)
    errExit("getsockopt");

  printf("Credentials from SO_PEERCRED: pid=%ld, euid=%ld, egid=%ld\n",
         (long) ucred.pid, (long) ucred.uid, (long) ucred.gid);
#endif

    exit(EXIT_SUCCESS);
}
