/************************************************************************* \
*                  Copyright (C) Michael Kerrisk, 2010.                   *
*                  Copyright (C) Svante Signell, 2013.                    *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* scm_righs+creds_send.c

   Used in conjunction with scm_rights+creds_recv.c to demonstrate
   passing of file descriptors and process credentials via a UNIX
   domain socket.

   This program sends file descriptors and credentials to a UNIX
   domain socket.

   Usage is as shown in the usageErr() call below.

   Credentials can exchanged over stream or datagram sockets. This program
   uses stream sockets by default; the "-d" command-line option specifies
   that datagram sockets should be used instead.

   Modified for kFreeBSD and Hurd by Svante Signell April-Aug 2013.
   Added -z option, send a zero byte for credentials with Hurd and kFreeBSD
   New file: scm_righs+creds_send.c, Svante Signell Nov 2013.
*/

#include "scm_rights+creds.h"

#ifndef SCM_CREDENTIALS
#define SCM_CREDENTIALS SCM_CREDS;
#endif

#ifndef NRIGHTS
#define NRIGHTS 1
#endif

#ifndef NCREDS
#define NCREDS 0
#endif

/* 
For Linux it does not matter if '\0' or something else is
sent. Credentials passing is controlled by setsockopt option
SO_PASSCRED used in scm_rigths+creds_recv.c
 */

int
main(int argc, char *argv[])
{
  struct msghdr msgh;
  struct iovec iov;
  int sfd, opt;

#if NRIGHTS == 1 || NRIGHTS == 3
      int fd1;
      char *filename1 = "./test1.txt";
#endif
#if NRIGHTS == 2 || NRIGHTS == 3
      int *fd2ptr = NULL, fd2[2];
      char *filename2 = "./test2.txt";
      char *filename3 = "./test3.txt";
#endif

  char zdata = {0};
  int data;
  ssize_t ns;
  Boolean useDatagramSocket, noExplicit, zerobyte;

#ifdef __linux__
  int sz =  sizeof(struct ucred);
#else
  int sz =  sizeof(struct cmsgcred);
#endif

  /* Space large enough to hold NCREDS ucred structures + NRIGHTS FDs */
  /* FIXME: Cases 0 does not work */

  int SW, cspc, cspc1, cspc2, cspc3;
  if (NRIGHTS == 0 && !NCREDS) SW = 0;
  if (NRIGHTS == 1 && !NCREDS) SW = 1;
  if (NRIGHTS == 2 && !NCREDS) SW = 2;
  if (NRIGHTS == 3 && !NCREDS) SW = 3;
  if (NRIGHTS == 0 && NCREDS) SW = 4;
  if (NRIGHTS == 1 && NCREDS) SW = 5;
  if (NRIGHTS == 2 && NCREDS) SW = 6;
  if (NRIGHTS == 3 && NCREDS) SW = 7;

  cspc1 = CMSG_SPACE(sizeof(int));
  cspc2 = CMSG_SPACE(sz);
  cspc3 = CMSG_SPACE(2 * sizeof(int));
  switch (SW)
    {
    case 0:
      cspc = CMSG_SPACE(sizeof (int));
      fprintf(stderr,"Case 1: cspc = %c [char]\n", cspc);
      break;
    case 1:
      cspc = cspc1;
      fprintf(stderr,"Case 1: cspc = %d\n", cspc);
      break;
    case 2:
      cspc = cspc3;
      fprintf(stderr,"Case 2: cspc = %d\n", cspc);
      break;
    case 3:
      cspc = cspc1 + cspc3;
      fprintf(stderr,"Case 3: cspc = %d\n", cspc);
      break;
    case 4:
      cspc = cspc2;
      fprintf(stderr,"Case 4: cspc = %d\n", cspc);
      break;
    case 5:
      cspc = cspc1 + cspc2;
      fprintf(stderr,"Case 5: cspc = %d\n", cspc);
      break;
    case 6:
      cspc = cspc2 + cspc3;
      fprintf(stderr,"Case 6: cspc = %d\n", cspc);
      break;
    case 7:
      cspc = cspc1 + cspc2 + cspc3;
      fprintf(stderr,"Case 7: cspc = %d\n", cspc);
      break;
    default:
      break;
    }

#if NRIGHTS !=0 || NCREDS !=0
  union {
  struct cmsghdr cmh;
  char control[cspc];
} control_un;
  struct cmsghdr *cmhp = NULL;
#endif

  /* Parse command-line arguments */
  useDatagramSocket = FALSE;
  noExplicit = FALSE;
  zerobyte = FALSE;

  while ((opt = getopt(argc, argv, "dnz")) != -1) {
    switch (opt) {
    case 'd':
      useDatagramSocket = TRUE;
      break;

    case 'n':
      noExplicit = TRUE;
      break;

    case 'z':
      zerobyte = TRUE;
      break;

    default:
#ifdef __linux__
      usageErr("%s [-d] [-n] [-z] [data [PID [UID [GID]]]]\n"
	       "        -d    use datagram socket\n"
	       "        -n    don't construct explicit credentials structure\n"
	       "        -z    send zero credentials byte\n", argv[0]);
#else
      usageErr("%s [-d] [-n] [-z] [data [PID [UID [EUID [GID]]]]]\n"
	       "        -d    use datagram socket\n"
	       "        -n    don't construct explicit credentials structure\n"
	       "        -z    send zero credentials byte\n", argv[0]);
#endif
    }
  }

  /* On Linux, we must transmit at least 1 byte of real data in
     order to send ancillary data */

  msgh.msg_iov = &iov;
  msgh.msg_iovlen = 1;

  /* Data is optionally taken from command line */

  if (zerobyte)
    {
      /* Send a zero credentials byte */
      zdata = '\0';
      iov.iov_base = &zdata;
      iov.iov_len = sizeof(char);
      printf("Sending a zero credentials byte to receive the credentials\n");
      printf("Sent data = '0x%0x'\n", zdata);
    }
  else
    {
      /* Send default integer 12345 or taken from the command line */
      data = (argc > optind) ? atoi(argv[optind]) : 12345;
      iov.iov_base = &data;
      iov.iov_len = sizeof(int);
      printf("Sent data = %d\n", data);
    }

  /* Don't need to specify destination address, because we use
     connect() below */
  msgh.msg_name = NULL;
  msgh.msg_namelen = 0;

  if (noExplicit)
    {
    /* Don't construct an explicit credentials structure. (It
       is not necessary to do so, if we just want the receiver to
       receive our real credentials.) */
    printf("Not explicitly sending a credentials structure\n");
    msgh.msg_control = NULL;
    msgh.msg_controllen = 0;
    } 

  if (!noExplicit)
    {
#if NRIGHTS !=0 || NCREDS !=0
      /* Set 'msgh' fields to describe 'control_un' */
      msgh.msg_control = control_un.control;
      msgh.msg_controllen = sizeof(control_un.control);
      fprintf(stderr,"msgh.msg_controllen = %d\n", (int)msgh.msg_controllen);
      /* http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=514917 */
      /* FIXME: CMSG_NXTHDR(&msgh, cmhp) returns NULL if not zeroed out */
      memset (msgh.msg_control, 0, msgh.msg_controllen);
#endif

#if NRIGHTS == 1 || NRIGHTS == 3
      /* Set message header to describe ancillary data (one FDs) that
	 we want to send */
      fd1 = open(filename1, O_RDONLY);
      if (fd1 == -1)
	errExit("open fd1");

      cmhp = CMSG_FIRSTHDR(&msgh);
      cmhp->cmsg_len = CMSG_LEN(sizeof(int));
      cmhp->cmsg_level = SOL_SOCKET;
      cmhp->cmsg_type = SCM_RIGHTS;
      *((int *) CMSG_DATA(cmhp)) = fd1;

      fprintf(stderr, "Sent file descriptor: fd1 = %d\n", fd1);
      fprintf(stderr, "cmhp->cmsg_len = %d\n", (int)cmhp->cmsg_len);
#endif /* NRIGHTS */

#if NCREDS !=0
    /* Construct an explicit credentials structure */
#ifdef __linux__
      /* NOTE: ucred does not have euid defined in the struct */
      struct ucred *ucp;
#else
      struct cmsgcred *ucp;
#endif
      /* Set message header to describe ancillary data that we want to send */
#if NCREDS != 0 && (NRIGHTS == 0 || NRIGHTS == 2)
      cmhp = CMSG_FIRSTHDR(&msgh);
#else
      cmhp = CMSG_NXTHDR(&msgh, cmhp);
#endif
      if (cmhp == NULL)
	errExit("cmhp == NULL");
      cmhp->cmsg_level = SOL_SOCKET;
#ifdef __linux__
      cmhp->cmsg_len = CMSG_LEN(sizeof(struct ucred));
      cmhp->cmsg_type = SCM_CREDENTIALS;
      ucp = (struct ucred *) CMSG_DATA(cmhp);
#else
      cmhp->cmsg_len = CMSG_LEN(sizeof(struct cmsgcred));
      cmhp->cmsg_type = SCM_CREDS;
      ucp = (struct cmsgcred *) CMSG_DATA(cmhp);
#endif

      /* Use sender's own PID, real UID, and real GID, unless
	 alternate values were supplied on the command line */
#ifdef __linux__
      ucp->pid = (argc > optind + 1 && strcmp(argv[optind + 1], "-") != 0) ?
	atoi(argv[optind + 1]) : getpid();
      ucp->uid = (argc > optind + 2 && strcmp(argv[optind + 2], "-") != 0) ?
	atoi(argv[optind + 2]) : getuid();
      ucp->gid = (argc > optind + 3 && strcmp(argv[optind + 3], "-") != 0) ?
	atoi(argv[optind + 3]) : getgid();
      printf("Sent credentials: pid=%ld, uid=%ld, euid=%d (from geteuid), gid=%ld\n",
	     (long) ucp->pid, (long) ucp->uid, geteuid(), (long) ucp->gid);
#else
      ucp->cmcred_pid = (argc > optind + 1 && strcmp(argv[optind + 1], "-") != 0) ?
	atoi(argv[optind + 1]) : getpid();
      ucp->cmcred_uid = (argc > optind + 2 && strcmp(argv[optind + 2], "-") != 0) ?
	atoi(argv[optind + 2]) : getuid();
      ucp->cmcred_euid = (argc > optind + 3 && strcmp(argv[optind + 3], "-") != 0) ?
	atoi(argv[optind + 3]) : geteuid();
      ucp->cmcred_gid = (argc > optind + 4 && strcmp(argv[optind + 4], "-") != 0) ?
	atoi(argv[optind + 4]) : getgid();
      printf("Sent credentials: cmcred_pid=%ld, cmcred_uid=%ld, cmcred_euid=%ld, cmcred_gid=%ld\n",
	     (long) ucp->cmcred_pid, (long) ucp->cmcred_uid, (long) ucp->cmcred_euid, (long) ucp->cmcred_gid);
#endif
      fprintf(stderr, "cmhp->cmsg_len = %d\n", (int)cmhp->cmsg_len);
#endif /* NCREDS */

      /* Set message header to describe ancillary data (two FDs) that
	 we want to send: see man cmsg */
#if NRIGHTS == 2 || NRIGHTS == 3
      fd2[0] = open(filename2, O_RDONLY);
      if (fd2[0] == -1)
	errExit("open fd2[0]");
      fd2[1] = open(filename3, O_RDONLY);
      if (fd2[1] == -1)
	errExit("open fd2[1]");

#if NRIGHTS == 2 && !NCREDS
      cmhp = CMSG_FIRSTHDR(&msgh);
#else
      cmhp = CMSG_NXTHDR(&msgh, cmhp);
#endif
      if (cmhp == NULL)
	errExit("cmhp == NULL");
#define NFD 2
      cmhp->cmsg_len = CMSG_LEN(NFD*sizeof(int));
      cmhp->cmsg_level = SOL_SOCKET;
      cmhp->cmsg_type = SCM_RIGHTS;
      fd2ptr = (int *) CMSG_DATA(cmhp);
      memcpy(fd2ptr, fd2, NFD*sizeof(int));
      fprintf(stderr, "Sent file descriptor: fd2[0]=%d, fd2[1]=%d\n", fd2[0], fd2[1]);
      fprintf(stderr, "cmhp->cmsg_len = %d\n", (int)cmhp->cmsg_len);
#endif /* NRIGHTS */
    } /* noExplicit */

  sfd = unixConnect(SOCK_PATH, useDatagramSocket ? SOCK_DGRAM : SOCK_STREAM);
  if (useDatagramSocket) {
    printf("Sending via datagram socket\n");
  }
  else
    {
      printf("Sending via stream socket\n");
    }
  if (sfd == -1)
    errExit("unixConnect");

  ns = sendmsg(sfd, &msgh, 0);
  if (ns == -1)
    errExit("sendmsg");
  printf("sendmsg() sent %ld bytes\n\n", (long) ns);
  unlink(SOCK_PATH);

  exit(EXIT_SUCCESS);
}
